@@ -17,17 +17,20 @@
 import json
 from datetime import timedelta
 
+import pyparsing as pp
+
 
 class MetricsReporter(object):
     """Generate metrics about events in a repository."""
 
     def __init__(self, repository):
         """Create a new instance.
-        
+
         :param repository: the repository to work with
         """
 
         self._repo = repository
+        self._selector_parser = None
 
     def get_report(self, selectors, start, end, period=None):
         selectors = self._expand_wildcard_selectors(selectors)
@@ -40,8 +43,9 @@ class MetricsReporter(object):
 
             while not done:
                 period_end = start + period - timedelta(days=1)
-                reports.append(self._get_report_for_period(selectors, start,
-                                                           period_end))
+                reports.append(
+                    self._get_report_for_period(selectors, start, period_end)
+                )
 
                 if period_end > end:
                     done = True
@@ -63,31 +67,36 @@ class MetricsReporter(object):
     def get_single_report(self, item_filter):
         """Get a single report object based on the given filter."""
 
-        issues_opened = [i for i in self._repo.all_issues
-                         if item_filter.filterIssue(i)]
-        issues_closes = [e for e in self._repo.events
-                         if e.event == 'closed'
-                         and not hasattr(e.issue, 'pull_request')
-                         and item_filter.filterEvent(e)]
+        issues_opened = [i for i in self._repo.all_issues if item_filter.filterIssue(i)]
+        issues_closes = [
+            e
+            for e in self._repo.events
+            if e.event == 'closed'
+            and not hasattr(e.issue, 'pull_request')
+            and item_filter.filterEvent(e)
+        ]
 
-        pulls_opened = [p for p in self._repo.all_pull_requests
-                        if item_filter.filterIssue(p)]
-        pulls_closes = [e for e in self._repo.events
-                        if e.event == 'closed'
-                        and hasattr(e.issue, 'pull_request')
-                        and item_filter.filterEvent(e)]
-        pulls_merged = [e for e in self._repo.events
-                        if e.event == 'merged'
-                        and item_filter.filterEvent(e)]
+        pulls_opened = [
+            p for p in self._repo.all_pull_requests if item_filter.filterIssue(p)
+        ]
+        pulls_closes = [
+            e
+            for e in self._repo.events
+            if e.event == 'closed'
+            and hasattr(e.issue, 'pull_request')
+            and item_filter.filterEvent(e)
+        ]
+        pulls_merged = [
+            e
+            for e in self._repo.events
+            if e.event == 'merged' and item_filter.filterEvent(e)
+        ]
 
-        comments = [c for c in self._repo.comments
-                    if item_filter.filterComment(c)]
+        comments = [c for c in self._repo.comments if item_filter.filterComment(c)]
 
-        commits = [c for c in self._repo.commits
-                   if item_filter.filterCommit(c)]
+        commits = [c for c in self._repo.commits if item_filter.filterCommit(c)]
 
-        releases = [r for r in self._repo.releases
-                    if item_filter.filterRelease(r)]
+        releases = [r for r in self._repo.releases if item_filter.filterRelease(r)]
 
         contributors = []
         contributors.extend([i.user.login for i in issues_opened])
@@ -97,16 +106,21 @@ class MetricsReporter(object):
         contributors.extend([e.actor.login for e in pulls_closes])
         contributors = set(contributors)
 
-        return _Report(str(item_filter), item_filter.name, [
-            len(issues_opened),
-            len(issues_closes),
-            len(pulls_opened),
-            len(pulls_closes),
-            len(pulls_merged),
-            len(comments),
-            len(commits),
-            len(releases),
-            len(contributors)])
+        return _Report(
+            str(item_filter),
+            item_filter.name,
+            [
+                len(issues_opened),
+                len(issues_closes),
+                len(pulls_opened),
+                len(pulls_closes),
+                len(pulls_merged),
+                len(comments),
+                len(commits),
+                len(releases),
+                len(contributors),
+            ],
+        )
 
     def _expand_wildcard_selectors(self, selectors):
         if not True in ['*' in s for s in selectors]:
@@ -114,46 +128,101 @@ class MetricsReporter(object):
 
         expanded_selectors = []
         for selector in selectors:
-            if selector == 'user:*':
-                expanded_selectors.extend([f'user:{u}' for u in self._repo.contributors])
-            elif selector == 'label:*':
-                expanded_selectors.extend([f'label:{l}' for l in self._repo.labels])
+            if 'user:*' in selector:
+                expanded_selectors.extend(
+                    [
+                        selector.replace('user:*', f'user:{u}')
+                        for u in self._repo.contributors
+                    ]
+                )
+            elif 'label:*' in selector:
+                expanded_selectors.extend(
+                    [
+                        selector.replace('label:*', f'label:{l}')
+                        for l in self._repo.labels
+                    ]
+                )
             else:
                 expanded_selectors.append(selector)
 
         return expanded_selectors
 
-    def _get_filter_from_selector(self, selector, level=0):
-        if ' AS ' in selector:
-            selector, name = selector.split(' AS ')
-        else:
-            name = selector
+    def _get_filter_from_selector(self, selector):
+        return self._get_parser().parse_string(selector).as_list()[0]
 
-        if selector[0] == '!':
-            f = ComplementFilter(self._get_filter_from_selector(selector[1:], level + 1))
-        elif selector == 'all':
-            f = NullFilter()
-        elif selector.startswith('team:'):
-            team_slug = selector[5:]
-            members = [m.login for m in self._repo.get_team(team_slug)]
-            f = TeamFilter(team_slug, members)
-        elif selector.startswith('user:'):
-            user_name = selector[5:]
-            f = UserFilter(user_name)
-            if selector == name:
-                name = f'@{user_name}'
-        elif selector.startswith('label:'):
-            label = selector[6:]
-            f = LabelFilter(label)
-            if selector == name:
-                name = label
-        else:
-            f = NullFilter()
+    def _get_parser(self):
+        if self._selector_parser is None:
+            filter_value = pp.Combine(
+                pp.Word(pp.alphanums + '-_')
+                + (pp.White() + pp.Word(pp.alphanums + '-_'))[...]
+            ).leaveWhitespace()
+            team_filter = (
+                (pp.Literal('team:') + filter_value)
+                .set_parse_action(
+                    lambda t: TeamFilter(
+                        t[1], [m.login for m in self._repo.get_team(t[1])]
+                    )
+                )
+                .leave_whitespace()
+            )
+            user_filter = (
+                (pp.Literal('user:') + filter_value)
+                .set_parse_action(lambda t: UserFilter(t[1]))
+                .leave_whitespace()
+            )
+            label_filter = (
+                (pp.Literal('label:') + filter_value)
+                .set_parse_action(lambda t: LabelFilter(t[1]))
+                .leave_whitespace()
+            )
+            all_filter = pp.Literal('all').set_parse_action(lambda _: NullFilter())
+            filter_item = all_filter | team_filter | user_filter | label_filter
 
-        if level == 0:
-            return CombinedFilter(name, [self._date_filter, f])
+            expression = pp.Forward()
+            complement_filter = (
+                (pp.Literal("!") + expression)
+                .set_parse_action(lambda t: ComplementFilter(t[1]))
+                .leave_whitespace()
+            )
+            atom = (
+                filter_item
+                | complement_filter
+                | (pp.Literal('(').suppress() + expression + pp.Literal(')').suppress())
+            )
+            operator = pp.one_of(['&', '|', '^'])
+            expression <<= (atom + operator + pp.White().suppress())[0, 1] + atom
+            expression.set_parse_action(self._expression_action)
+
+            selector = (
+                expression
+                + (
+                    (pp.Literal('AS') | pp.Literal('=')).suppress() + pp.Word(pp.alphas)
+                )[0, 1]
+                + pp.StringEnd()
+            ).set_parse_action(self._selector_action)
+            self._selector_parser = selector
+        return self._selector_parser
+
+    def _selector_action(self, tokens):
+        if len(tokens) == 2:
+            name = tokens[1]
         else:
-            return f
+            name = tokens[0].name
+        return NamedFilter(name, [self._date_filter, tokens[0]])
+
+    def _expression_action(self, tokens):
+        if len(tokens) == 3:
+            if tokens[1] == '&':
+                return IntersectionFilter([tokens[0], tokens[2]])
+            elif tokens[1] == '|':
+                return UnionFilter([tokens[0], tokens[2]])
+            elif tokens[1] == '^':
+                return DifferenceFilter([tokens[0], tokens[2]])
+            else:
+                # should not happen
+                return NullFilter()
+        else:
+            return tokens[0]
 
 
 class MetricsFormatter(object):
@@ -176,7 +245,6 @@ class MetricsFormatter(object):
 
 
 class JsonMetricsFormatter(MetricsFormatter):
-
     def write(self, metrics, output):
         if isinstance(metrics, list):
             d = [self._get_dict_for_period(m) for m in metrics]
@@ -188,10 +256,10 @@ class JsonMetricsFormatter(MetricsFormatter):
         d = {
             'period': {
                 'to': metrics.end_date.strftime('%Y-%m-%d'),
-                'from': metrics.start_date.strftime('%Y-%m-%d')
-                },
-            'contributions': []
-            }
+                'from': metrics.start_date.strftime('%Y-%m-%d'),
+            },
+            'contributions': [],
+        }
 
         for report in metrics.contributions:
             c = {
@@ -200,25 +268,24 @@ class JsonMetricsFormatter(MetricsFormatter):
                     'contributors': report.contributors,
                     'issues': {
                         'opened': report.issues_opened,
-                        'closed': report.issues_closed
-                        },
+                        'closed': report.issues_closed,
+                    },
                     'pull_requests': {
                         'opened': report.pull_requests_opened,
                         'closed': report.pull_requests_closed,
-                        'merged': report.pull_requests_merged
-                        },
+                        'merged': report.pull_requests_merged,
+                    },
                     'comments': report.comments,
                     'commits': report.commits,
-                    'releases': report.releases
-                    }
-                }
+                    'releases': report.releases,
+                },
+            }
             d['contributions'].append(c)
 
         return d
 
 
 class MarkdownMetricsFormatter(MetricsFormatter):
-
     def write(self, reportset, output):
         if isinstance(reportset, list):
             for r in reportset:
@@ -263,8 +330,8 @@ class MarkdownMetricsFormatter(MetricsFormatter):
             "Pull requests merged",
             "Comments",
             "Commits",
-            "Releases"
-            ]
+            "Releases",
+        ]
 
         for item in items:
             self._write_line(item, reportset.contributions, output, with_total)
@@ -293,16 +360,24 @@ class MarkdownMetricsFormatter(MetricsFormatter):
 
 
 class CsvMetricsFormatter(MetricsFormatter):
-
     def __init__(self, sep=','):
         self._sep = sep
 
     def write(self, reportset, output):
-        headers = ['Date', 'Selector', 'Selector name',
-                   'Issues opened', 'Issues closed',
-                   'Pull requests opened', 'Pull requests closed',
-                   'Pull requests merged', 'Comments', 'Commits',
-                   'Releases', 'Contributors']
+        headers = [
+            'Date',
+            'Selector',
+            'Selector name',
+            'Issues opened',
+            'Issues closed',
+            'Pull requests opened',
+            'Pull requests closed',
+            'Pull requests merged',
+            'Comments',
+            'Commits',
+            'Releases',
+            'Contributors',
+        ]
         output.write(self._sep.join(headers))
         output.write('\n')
 
@@ -314,12 +389,20 @@ class CsvMetricsFormatter(MetricsFormatter):
 
     def _write_reportset(self, reportset, output):
         for report in reportset.contributions:
-            values = [reportset.end_date.strftime('%Y-%m-%d'),
-                      report.selector, report.name,
-                      report.issues_opened, report.issues_closed,
-                      report.pull_requests_opened, report.pull_requests_closed,
-                      report.pull_requests_merged, report.comments,
-                      report.commits, report.releases, report.contributors]
+            values = [
+                reportset.end_date.strftime('%Y-%m-%d'),
+                report.selector,
+                report.name,
+                report.issues_opened,
+                report.issues_closed,
+                report.pull_requests_opened,
+                report.pull_requests_closed,
+                report.pull_requests_merged,
+                report.comments,
+                report.commits,
+                report.releases,
+                report.contributors,
+            ]
             output.write(self._sep.join([str(v) for v in values]))
             output.write('\n')
 
@@ -407,7 +490,6 @@ class _Report(object):
 
 
 class ItemFilter(object):
-
     def filterIssue(self, issue):
         return self._filterItem(issue)
 
@@ -432,7 +514,6 @@ class ItemFilter(object):
 
 
 class NullFilter(ItemFilter):
-
     def _filterItem(self, _):
         return True
 
@@ -441,7 +522,6 @@ class NullFilter(ItemFilter):
 
 
 class DateRangeFilter(ItemFilter):
-
     def __init__(self, start, end):
         self._start = start
         self._end = end
@@ -454,7 +534,6 @@ class DateRangeFilter(ItemFilter):
 
 
 class TeamFilter(ItemFilter):
-
     def __init__(self, team_name, members):
         self._members = members
         self._slug = team_name
@@ -462,6 +541,10 @@ class TeamFilter(ItemFilter):
     def __str__(self):
         return f'team:{self._slug}'
 
+    @property
+    def name(self):
+        return self._slug
+
     def filterIssue(self, issue):
         return issue.user.login in self._members
 
@@ -482,22 +565,28 @@ class TeamFilter(ItemFilter):
 
 
 class UserFilter(TeamFilter):
-
     def __init__(self, user_name):
         TeamFilter.__init__(self, None, [user_name])
 
     def __str__(self):
         return f'user:{self._members[0]}'
 
+    @property
+    def name(self):
+        return f'@{self._members[0]}'
+
 
 class LabelFilter(ItemFilter):
-
     def __init__(self, label):
         self._label = label
 
     def __str__(self):
         return f'label:{self._label}'
 
+    @property
+    def name(self):
+        return self._label
+
     def filterIssue(self, issue):
         return self._label in [l.name for l in issue.labels]
 
@@ -506,13 +595,16 @@ class LabelFilter(ItemFilter):
 
 
 class ComplementFilter(ItemFilter):
-
     def __init__(self, inner_filter):
         self._filter = inner_filter
 
     def __str__(self):
         return f'!{str(self._filter)}'
 
+    @property
+    def name(self):
+        return f'!{self._filter.name}'
+
     def filterIssue(self, issue):
         return not self._filter.filterIssue(issue)
 
@@ -530,31 +622,69 @@ class ComplementFilter(ItemFilter):
 
 
 class CombinedFilter(ItemFilter):
-
-    def __init__(self, name, filters):
-        self._name = name
+    def __init__(self, filters, operator):
         self._filters = filters
+        self._op = operator
 
     def __str__(self):
-        components = [str(f) for f in self._filters if not str(f).startswith('date:')]
-        return ' & '.join(components)
+        return f' {self._op} '.join([str(f) for f in self._filters])
+
+    @property
+    def name(self):
+        return f' {self._op } '.join([f.name for f in self._filters])
+
+    def filterIssue(self, issue):
+        return self._apply([f.filterIssue(issue) for f in self._filters])
+
+    def filterEvent(self, event):
+        return self._apply([f.filterEvent(event) for f in self._filters])
+
+    def filterComment(self, comment):
+        return self._apply([f.filterComment(comment) for f in self._filters])
+
+    def filterCommit(self, commit):
+        return self._apply([f.filterCommit(commit) for f in self._filters])
+
+    def filterRelease(self, release):
+        return self._apply([f.filterRelease(release) for f in self._filters])
+
+    def _apply(self, _):
+        pass
+
+
+class IntersectionFilter(CombinedFilter):
+    def __init__(self, filters):
+        CombinedFilter.__init__(self, filters, '&')
+
+    def _apply(self, results):
+        return not False in results
+
+
+class UnionFilter(CombinedFilter):
+    def __init__(self, filters):
+        CombinedFilter.__init__(self, filters, '|')
+
+    def _apply(self, results):
+        return True in results
+
+
+class DifferenceFilter(CombinedFilter):
+    def __init__(self, filters):
+        CombinedFilter.__init__(self, filters, '^')
+
+    def _apply(self, results):
+        return len([r for r in results if r == True]) == 1
+
+
+class NamedFilter(IntersectionFilter):
+    def __init__(self, name, filters):
+        IntersectionFilter.__init__(self, filters)
+        self._name = name
 
     @property
     def name(self):
         return self._name
 
-    def filterIssue(self, issue):
-        return not False in [f.filterIssue(issue) for f in self._filters]
-
-    def filterEvent(self, event):
-        return not False in [f.filterEvent(event) for f in self._filters]
-
-    def filterComment(self, comment):
-        return not False in [f.filterComment(comment) for f in self._filters]
-
-    def filterCommit(self, commit):
-        return not False in [f.filterCommit(commit) for f in self._filters]
-
-    def filterRelease(self, release):
-        return not False in [f.filterRelease(release) for f in self._filters]
-
+    def __str__(self):
+        components = [str(f) for f in self._filters if not str(f).startswith('date:')]
+        return ' & '.join(components)
